#![allow(clippy::borrow_interior_mutable_const, clippy::type_complexity)]

//! Static files support
use std::cell::RefCell;
use std::path::PathBuf;
use std::rc::Rc;
use std::sync::{Arc, RwLock};

use actix_service::boxed::{self, BoxServiceFactory};
use actix_service::{IntoServiceFactory, ServiceFactory};
use actix_web::dev::{
  AppService, HttpServiceFactory, ResourceDef, ServiceRequest, ServiceResponse,
};
use actix_web::error::Error as ActixError;
use actix_web::guard::Guard;
use futures_util::future::{ok, FutureExt, LocalBoxFuture};

use path_context::PathContext;
use service::FilesService;

mod chunked;
pub mod directory;
pub mod named;
mod pages;
mod path;
pub mod path_context;
mod pathbuf;
pub mod range;
mod service;

type HttpNewService = BoxServiceFactory<(), ServiceRequest, ServiceResponse, ActixError, ()>;

/// Static files handling
///
/// `Files` service must be registered with `App::service()` method.
///
/// ```rust
/// use std::{
///   convert::TryInto,
///   sync::{Arc, RwLock},
/// };
/// use actix_web::App;
/// use espresso::{
///   files::Files,
///   config::ServerConfig,
/// };
///
/// let mut server_config = ServerConfig::default();
///
/// let serve_dir = Arc::new(RwLock::new(Some(".".into())));
/// let root_path_context = Arc::new((&server_config).try_into().unwrap());
/// let path_contexts = Arc::new(vec![]);
///
/// let app = App::new()
///   .service(Files::new(
///     "/static",
///     serve_dir,
///     root_path_context,
///     path_contexts,
///   ));
/// ```
pub struct Files {
  path: String,
  directory: Arc<RwLock<Option<PathBuf>>>,
  root_path_context: Arc<PathContext>,
  #[allow(clippy::rc_buffer)]
  path_contexts: Arc<Vec<PathContext>>,
  redirect_to_slash: bool,
  default: Rc<RefCell<Option<Rc<HttpNewService>>>>,
  guards: Option<Rc<dyn Guard>>,
}

impl Clone for Files {
  fn clone(&self) -> Self {
    Self {
      directory: self.directory.clone(),
      redirect_to_slash: self.redirect_to_slash,
      default: self.default.clone(),
      path: self.path.clone(),
      guards: self.guards.clone(),
      root_path_context: self.root_path_context.clone(),
      path_contexts: self.path_contexts.clone(),
    }
  }
}

impl Files {
  /// Create new `Files` instance for specified base directory.
  ///
  /// `File` uses `ThreadPool` for blocking filesystem operations.
  /// By default pool with 5x threads of available cpus is used.
  /// Pool size can be changed by setting ACTIX_THREADPOOL environment variable.
  #[allow(clippy::rc_buffer)]
  pub fn new(
    path: &str,
    directory: Arc<RwLock<Option<PathBuf>>>,
    root_path_context: Arc<PathContext>,
    path_contexts: Arc<Vec<PathContext>>,
  ) -> Files {
    // let orig_dir = dir.into();
    // let dir = match orig_dir.canonicalize() {
    //   Ok(canon_dir) => canon_dir,
    //   Err(_) => {
    //     log::error!("Specified path is not a directory: {:?}", orig_dir);
    //     PathBuf::new()
    //   }
    // };

    Files {
      path: path.to_string(),
      directory,
      redirect_to_slash: false,
      default: Rc::new(RefCell::new(None)),
      guards: None,
      root_path_context,
      path_contexts,
    }
  }

  /// Redirects to a slash-ended path when browsing a directory.
  ///
  /// By default never redirect.
  pub fn redirect_to_slash_directory(mut self) -> Self {
    self.redirect_to_slash = true;
    self
  }

  /// Specifies custom guards to use for directory listings and files.
  ///
  /// Default behaviour allows GET and HEAD.
  #[inline]
  pub fn use_guards<G: Guard + 'static>(mut self, guards: G) -> Self {
    self.guards = Some(Rc::new(guards));
    self
  }

  /// Sets default handler which is used when no matched file could be found.
  pub fn default_handler<F, U>(mut self, f: F) -> Self
  where
    F: IntoServiceFactory<U>,
    U: ServiceFactory<
      Config = (),
      Request = ServiceRequest,
      Response = ServiceResponse,
      Error = ActixError,
    > + 'static,
  {
    // create and configure default resource
    self.default = Rc::new(RefCell::new(Some(Rc::new(boxed::factory(
      f.into_factory().map_init_err(|_| ()),
    )))));

    self
  }
}

impl HttpServiceFactory for Files {
  fn register(self, config: &mut AppService) {
    if self.default.borrow().is_none() {
      *self.default.borrow_mut() = Some(config.default_service());
    }

    let rdef = if config.is_root() {
      ResourceDef::root_prefix(&self.path)
    } else {
      ResourceDef::prefix(&self.path)
    };

    config.register_service(rdef, None, self, None)
  }
}

impl ServiceFactory for Files {
  type Request = ServiceRequest;
  type Response = ServiceResponse;
  type Error = ActixError;
  type Config = ();
  type Service = FilesService;
  type InitError = ();
  type Future = LocalBoxFuture<'static, Result<Self::Service, Self::InitError>>;

  fn new_service(&self, _: ()) -> Self::Future {
    let mut srv = FilesService::new(
      self.directory.clone(),
      self.redirect_to_slash,
      None,
      self.root_path_context.clone(),
      self.path_contexts.clone(),
      self.guards.clone(),
    );

    if let Some(ref default) = *self.default.borrow() {
      default
        .new_service(())
        .map(move |result| match result {
          Ok(default) => {
            srv.set_default(Some(default));
            Ok(srv)
          }
          Err(_) => Err(()),
        })
        .boxed_local()
    } else {
      ok(srv).boxed_local()
    }
  }
}

#[cfg(test)]
mod tests {
  use std::fs;
  use std::ops::Add;
  use std::{
    convert::TryInto,
    time::{Duration, SystemTime},
  };

  use super::*;
  use crate::config::{ContentDispositionConfig, IndexStrategyConfig, ServerConfig};
  use actix_web::guard;
  use actix_web::http::header::{self, ContentDisposition, DispositionParam, DispositionType};
  use actix_web::http::{Method, StatusCode};
  use actix_web::middleware::Compress;
  use actix_web::test::{self, TestRequest};
  use actix_web::{web, App, HttpResponse, Responder};
  use bytes::Bytes;
  use fs::File;
  use named::NamedFile;

  fn serve_dir<T: Into<PathBuf>>(path: T) -> Arc<RwLock<Option<PathBuf>>> {
    Arc::new(RwLock::new(Some(path.into())))
  }

  #[actix_rt::test]
  async fn test_if_modified_since_without_if_none_match() {
    let file = NamedFile::open("Cargo.toml").unwrap();
    let since = header::HttpDate::from(SystemTime::now().add(Duration::from_secs(60)));

    let req = TestRequest::default()
      .header(header::IF_MODIFIED_SINCE, since)
      .to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(resp.status(), StatusCode::NOT_MODIFIED);
  }

  #[actix_rt::test]
  async fn test_if_modified_since_with_if_none_match() {
    let file = NamedFile::open("Cargo.toml").unwrap();
    let since = header::HttpDate::from(SystemTime::now().add(Duration::from_secs(60)));

    let req = TestRequest::default()
      .header(header::IF_NONE_MATCH, "miss_etag")
      .header(header::IF_MODIFIED_SINCE, since)
      .to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_ne!(resp.status(), StatusCode::NOT_MODIFIED);
  }

  #[actix_rt::test]
  async fn test_named_file_text() {
    assert!(NamedFile::open("test--").is_err());
    let mut file = NamedFile::open("Cargo.toml").unwrap();
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "text/x-toml"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"Cargo.toml\""
    );
  }

  #[actix_rt::test]
  async fn test_named_file_content_disposition() {
    assert!(NamedFile::open("test--").is_err());
    let mut file = NamedFile::open("Cargo.toml").unwrap();
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"Cargo.toml\""
    );

    let file = NamedFile::open("Cargo.toml")
      .unwrap()
      .disable_content_disposition();
    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert!(resp.headers().get(header::CONTENT_DISPOSITION).is_none());
  }

  #[actix_rt::test]
  async fn test_named_file_non_ascii_file_name() {
    let mut file = NamedFile::from_file(File::open("Cargo.toml").unwrap(), "貨物.toml").unwrap();
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "text/x-toml"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"貨物.toml\"; filename*=UTF-8''%E8%B2%A8%E7%89%A9.toml"
    );
  }

  #[actix_rt::test]
  async fn test_named_file_set_content_type() {
    let mut file = NamedFile::open("Cargo.toml")
      .unwrap()
      .set_content_type(mime::TEXT_XML);
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "text/xml"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"Cargo.toml\""
    );
  }

  #[actix_rt::test]
  async fn test_named_file_image() {
    let mut file = NamedFile::open("tests/test.png").unwrap();
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "image/png"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"test.png\""
    );
  }

  #[actix_rt::test]
  async fn test_named_file_image_attachment() {
    let cd = ContentDisposition {
      disposition: DispositionType::Attachment,
      parameters: vec![DispositionParam::Filename(String::from("test.png"))],
    };
    let mut file = NamedFile::open("tests/test.png")
      .unwrap()
      .set_content_disposition(cd);
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "image/png"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "attachment; filename=\"test.png\""
    );
  }

  #[actix_rt::test]
  async fn test_named_file_binary() {
    let mut file = NamedFile::open("tests/test.binary").unwrap();
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "application/octet-stream"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "attachment; filename=\"test.binary\""
    );
  }

  #[actix_rt::test]
  async fn test_named_file_status_code_text() {
    let mut file = NamedFile::open("Cargo.toml")
      .unwrap()
      .set_status_code(StatusCode::NOT_FOUND);
    {
      file.file();
      let _f: &File = &file;
    }
    {
      let _f: &mut File = &mut file;
    }

    let req = TestRequest::default().to_http_request();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "text/x-toml"
    );
    assert_eq!(
      resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
      "inline; filename=\"Cargo.toml\""
    );
    assert_eq!(resp.status(), StatusCode::NOT_FOUND);
  }

  #[actix_rt::test]
  async fn test_mime_override() {
    let mut server_config = ServerConfig::default();

    server_config.index_strategy = Some(IndexStrategyConfig::IndexFiles {
      filenames: ["Cargo.toml".to_owned()].iter().cloned().collect(),
    });
    server_config.mime_disposition = Some(
      vec![(
        "text/x-toml".to_owned(),
        ContentDispositionConfig::Attachment,
      )]
      .iter()
      .cloned()
      .collect(),
    );

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;

    let request = TestRequest::get().uri("/").to_request();
    let response = test::call_service(&mut srv, request).await;
    assert_eq!(response.status(), StatusCode::OK);

    let content_disposition = response
      .headers()
      .get(header::CONTENT_DISPOSITION)
      .expect("To have CONTENT_DISPOSITION");
    let content_disposition = content_disposition
      .to_str()
      .expect("Convert CONTENT_DISPOSITION to str");
    assert_eq!(content_disposition, "attachment; filename=\"Cargo.toml\"");
  }

  #[actix_rt::test]
  async fn test_named_file_ranges_status_code() {
    let mut server_config = ServerConfig::default();

    server_config.index_strategy = Some(IndexStrategyConfig::IndexFiles {
      filenames: ["Cargo.toml".to_owned()].iter().cloned().collect(),
    });

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/test",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;

    // Valid range header
    let request = TestRequest::get()
      .uri("/t%65st/Cargo.toml")
      .header(header::RANGE, "bytes=10-20")
      .to_request();
    let response = test::call_service(&mut srv, request).await;
    assert_eq!(response.status(), StatusCode::PARTIAL_CONTENT);

    // Invalid range header
    let request = TestRequest::get()
      .uri("/t%65st/Cargo.toml")
      .header(header::RANGE, "bytes=1-0")
      .to_request();
    let response = test::call_service(&mut srv, request).await;

    assert_eq!(response.status(), StatusCode::RANGE_NOT_SATISFIABLE);
  }

  #[actix_rt::test]
  async fn test_named_file_content_range_headers() {
    let srv = test::start(|| {
      let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
      let path_contexts = Arc::new(vec![]);

      App::new().service(Files::new(
        "/",
        serve_dir("."),
        root_path_context,
        path_contexts,
      ))
    });

    // Valid range header
    let response = srv
      .get("/tests/test.binary")
      .header(header::RANGE, "bytes=10-20")
      .send()
      .await
      .unwrap();
    let content_range = response.headers().get(header::CONTENT_RANGE).unwrap();
    assert_eq!(content_range.to_str().unwrap(), "bytes 10-20/100");

    // Invalid range header
    let response = srv
      .get("/tests/test.binary")
      .header(header::RANGE, "bytes=10-5")
      .send()
      .await
      .unwrap();
    let content_range = response.headers().get(header::CONTENT_RANGE).unwrap();
    assert_eq!(content_range.to_str().unwrap(), "bytes */100");
  }

  #[actix_rt::test]
  async fn test_named_file_content_length_headers() {
    let srv = test::start(|| {
      let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
      let path_contexts = Arc::new(vec![]);

      App::new().service(Files::new(
        "/",
        serve_dir("."),
        root_path_context,
        path_contexts,
      ))
    });

    // Valid range header
    let response = srv
      .get("/tests/test.binary")
      .header(header::RANGE, "bytes=10-20")
      .send()
      .await
      .unwrap();
    let content_length = response.headers().get(header::CONTENT_LENGTH).unwrap();
    assert_eq!(content_length.to_str().unwrap(), "11");

    // Valid range header, starting from 0
    let response = srv
      .get("/tests/test.binary")
      .header(header::RANGE, "bytes=0-20")
      .send()
      .await
      .unwrap();
    let content_length = response.headers().get(header::CONTENT_LENGTH).unwrap();
    assert_eq!(content_length.to_str().unwrap(), "21");

    // Without range header
    let mut response = srv.get("/tests/test.binary").send().await.unwrap();
    let content_length = response.headers().get(header::CONTENT_LENGTH).unwrap();
    assert_eq!(content_length.to_str().unwrap(), "100");

    // Should be no transfer-encoding
    let transfer_encoding = response.headers().get(header::TRANSFER_ENCODING);
    assert!(transfer_encoding.is_none());

    // Check file contents
    let bytes = response.body().await.unwrap();
    let data = Bytes::from(fs::read("tests/test.binary").unwrap());
    assert_eq!(bytes, data);
  }

  #[actix_rt::test]
  async fn test_head_content_length_headers() {
    let srv = test::start(|| {
      let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
      let path_contexts = Arc::new(vec![]);

      App::new().service(Files::new(
        "/",
        serve_dir("."),
        root_path_context,
        path_contexts,
      ))
    });

    let response = srv.head("/tests/test.binary").send().await.unwrap();

    let content_length = response
      .headers()
      .get(header::CONTENT_LENGTH)
      .unwrap()
      .to_str()
      .unwrap();

    assert_eq!(content_length, "100");
  }

  #[actix_rt::test]
  async fn test_static_files_with_spaces() {
    let mut server_config = ServerConfig::default();

    server_config.index_strategy = Some(IndexStrategyConfig::IndexFiles {
      filenames: ["Cargo.toml".to_owned()].iter().cloned().collect(),
    });

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;
    let request = TestRequest::get()
      .uri("/tests/test%20space.binary")
      .to_request();
    let response = test::call_service(&mut srv, request).await;
    assert_eq!(response.status(), StatusCode::OK);

    let bytes = test::read_body(response).await;
    let data = Bytes::from(fs::read("tests/test space.binary").unwrap());
    assert_eq!(bytes, data);
  }

  #[actix_rt::test]
  async fn test_files_not_allowed() {
    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().default_service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;

    let req = TestRequest::default()
      .uri("/Cargo.toml")
      .method(Method::POST)
      .to_request();

    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::METHOD_NOT_ALLOWED);

    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().default_service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;
    let req = TestRequest::default()
      .method(Method::PUT)
      .uri("/Cargo.toml")
      .to_request();
    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::METHOD_NOT_ALLOWED);
  }

  #[actix_rt::test]
  async fn test_files_guards() {
    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(
      Files::new("/", serve_dir("."), root_path_context, path_contexts).use_guards(guard::Post()),
    ))
    .await;

    let req = TestRequest::default()
      .uri("/Cargo.toml")
      .method(Method::POST)
      .to_request();

    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::OK);
  }

  #[actix_rt::test]
  async fn test_named_file_content_encoding() {
    let mut srv = test::init_service(App::new().wrap(Compress::default()).service(
      web::resource("/").to(|| async {
        NamedFile::open("Cargo.toml")
          .unwrap()
          .set_content_encoding(header::ContentEncoding::Identity)
      }),
    ))
    .await;

    let request = TestRequest::get()
      .uri("/")
      .header(header::ACCEPT_ENCODING, "gzip")
      .to_request();
    let res = test::call_service(&mut srv, request).await;
    assert_eq!(res.status(), StatusCode::OK);
    assert!(!res.headers().contains_key(header::CONTENT_ENCODING));
  }

  #[actix_rt::test]
  async fn test_named_file_content_encoding_gzip() {
    let mut srv = test::init_service(App::new().wrap(Compress::default()).service(
      web::resource("/").to(|| async {
        NamedFile::open("Cargo.toml")
          .unwrap()
          .set_content_encoding(header::ContentEncoding::Gzip)
      }),
    ))
    .await;

    let request = TestRequest::get()
      .uri("/")
      .header(header::ACCEPT_ENCODING, "gzip")
      .to_request();
    let res = test::call_service(&mut srv, request).await;
    assert_eq!(res.status(), StatusCode::OK);
    assert_eq!(
      res
        .headers()
        .get(header::CONTENT_ENCODING)
        .unwrap()
        .to_str()
        .unwrap(),
      "gzip"
    );
  }

  #[actix_rt::test]
  async fn test_named_file_allowed_method() {
    let req = TestRequest::default().method(Method::GET).to_http_request();
    let file = NamedFile::open("Cargo.toml").unwrap();
    let resp = file.respond_to(&req).await.unwrap();
    assert_eq!(resp.status(), StatusCode::OK);
  }

  #[actix_rt::test]
  async fn test_static_files() {
    let mut server_config = ServerConfig::default();

    server_config.index_strategy = Some(IndexStrategyConfig::AlwaysShowListing);

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;
    let req = TestRequest::with_uri("/missing").to_request();

    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::NOT_FOUND);

    // Without index strategy.
    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;

    let req = TestRequest::default().to_request();
    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::NOT_FOUND);

    // With default renderer.
    server_config.index_strategy = Some(IndexStrategyConfig::AlwaysShowListing);

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut srv = test::init_service(App::new().service(Files::new(
      "/",
      serve_dir("."),
      root_path_context,
      path_contexts,
    )))
    .await;
    let req = TestRequest::with_uri("/tests").to_request();
    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(
      resp.headers().get(header::CONTENT_TYPE).unwrap(),
      "text/html; charset=utf-8"
    );

    let bytes = test::read_body(resp).await;
    assert!(format!("{:?}", bytes).contains("/tests/test.png"));
  }

  #[actix_rt::test]
  async fn test_redirect_to_slash_directory() {
    let mut server_config = ServerConfig::default();

    server_config.index_strategy = Some(IndexStrategyConfig::IndexFiles {
      filenames: ["test.png".to_owned()].iter().cloned().collect(),
    });

    let root_path_context = Arc::new((&server_config).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    // should not redirect if no index
    // let mut srv = test::init_service(
    //   App::new().service(Files::new("/", serve_dir(".")).redirect_to_slash_directory()),
    // )
    // .await;
    // let req = TestRequest::with_uri("/tests").to_request();
    // let resp = test::call_service(&mut srv, req).await;
    // assert_eq!(resp.status(), StatusCode::NOT_FOUND);

    // should redirect if index present
    let mut srv = test::init_service(
      App::new().service(
        Files::new("/", serve_dir("."), root_path_context, path_contexts)
          .redirect_to_slash_directory(),
      ),
    )
    .await;
    let req = TestRequest::with_uri("/tests").to_request();
    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::FOUND);

    // should not redirect if the path is wrong
    let req = TestRequest::with_uri("/not_existing").to_request();
    let resp = test::call_service(&mut srv, req).await;
    assert_eq!(resp.status(), StatusCode::NOT_FOUND);
  }

  #[actix_rt::test]
  async fn test_static_files_bad_directory() {
    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let _st: Files = Files::new("/", serve_dir("missing"), root_path_context, path_contexts);

    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let _st: Files = Files::new(
      "/",
      serve_dir("Cargo.toml"),
      root_path_context,
      path_contexts,
    );
  }

  #[actix_rt::test]
  async fn test_default_handler_file_missing() {
    let root_path_context = Arc::new((&ServerConfig::default()).try_into().unwrap());
    let path_contexts = Arc::new(vec![]);

    let mut st = Files::new("/", serve_dir("."), root_path_context, path_contexts)
      .default_handler(|req: ServiceRequest| {
        ok(req.into_response(HttpResponse::Ok().body("default content")))
      })
      .new_service(())
      .await
      .unwrap();
    let req = TestRequest::with_uri("/missing").to_srv_request();

    let resp = test::call_service(&mut st, req).await;
    assert_eq!(resp.status(), StatusCode::OK);
    let bytes = test::read_body(resp).await;
    assert_eq!(bytes, Bytes::from_static(b"default content"));
  }

  //     #[actix_rt::test]
  //     async fn test_serve_index() {
  //         let st = Files::new(".").index_file("test.binary");
  //         let req = TestRequest::default().uri("/tests").finish();

  //         let resp = st.handle(&req).respond_to(&req).unwrap();
  //         let resp = resp.as_msg();
  //         assert_eq!(resp.status(), StatusCode::OK);
  //         assert_eq!(
  //             resp.headers()
  //                 .get(header::CONTENT_TYPE)
  //                 .expect("content type"),
  //             "application/octet-stream"
  //         );
  //         assert_eq!(
  //             resp.headers()
  //                 .get(header::CONTENT_DISPOSITION)
  //                 .expect("content disposition"),
  //             "attachment; filename=\"test.binary\""
  //         );

  //         let req = TestRequest::default().uri("/tests/").finish();
  //         let resp = st.handle(&req).respond_to(&req).unwrap();
  //         let resp = resp.as_msg();
  //         assert_eq!(resp.status(), StatusCode::OK);
  //         assert_eq!(
  //             resp.headers().get(header::CONTENT_TYPE).unwrap(),
  //             "application/octet-stream"
  //         );
  //         assert_eq!(
  //             resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
  //             "attachment; filename=\"test.binary\""
  //         );

  //         // nonexistent index file
  //         let req = TestRequest::default().uri("/tests/unknown").finish();
  //         let resp = st.handle(&req).respond_to(&req).unwrap();
  //         let resp = resp.as_msg();
  //         assert_eq!(resp.status(), StatusCode::NOT_FOUND);

  //         let req = TestRequest::default().uri("/tests/unknown/").finish();
  //         let resp = st.handle(&req).respond_to(&req).unwrap();
  //         let resp = resp.as_msg();
  //         assert_eq!(resp.status(), StatusCode::NOT_FOUND);
  //     }

  //     #[actix_rt::test]
  //     async fn test_serve_index_nested() {
  //         let st = Files::new(".").index_file("mod.rs");
  //         let req = TestRequest::default().uri("/src/client").finish();
  //         let resp = st.handle(&req).respond_to(&req).unwrap();
  //         let resp = resp.as_msg();
  //         assert_eq!(resp.status(), StatusCode::OK);
  //         assert_eq!(
  //             resp.headers().get(header::CONTENT_TYPE).unwrap(),
  //             "text/x-rust"
  //         );
  //         assert_eq!(
  //             resp.headers().get(header::CONTENT_DISPOSITION).unwrap(),
  //             "inline; filename=\"mod.rs\""
  //         );
  //     }

  //     #[actix_rt::test]
  //     fn integration_serve_index() {
  //         let mut srv = test::TestServer::with_factory(|| {
  //             App::new().handler(
  //                 "test",
  //                 Files::new(".").index_file("Cargo.toml"),
  //             )
  //         });

  //         let request = srv.get().uri(srv.url("/test")).finish().unwrap();
  //         let response = srv.execute(request.send()).unwrap();
  //         assert_eq!(response.status(), StatusCode::OK);
  //         let bytes = srv.execute(response.body()).unwrap();
  //         let data = Bytes::from(fs::read("Cargo.toml").unwrap());
  //         assert_eq!(bytes, data);

  //         let request = srv.get().uri(srv.url("/test/")).finish().unwrap();
  //         let response = srv.execute(request.send()).unwrap();
  //         assert_eq!(response.status(), StatusCode::OK);
  //         let bytes = srv.execute(response.body()).unwrap();
  //         let data = Bytes::from(fs::read("Cargo.toml").unwrap());
  //         assert_eq!(bytes, data);

  //         // nonexistent index file
  //         let request = srv.get().uri(srv.url("/test/unknown")).finish().unwrap();
  //         let response = srv.execute(request.send()).unwrap();
  //         assert_eq!(response.status(), StatusCode::NOT_FOUND);

  //         let request = srv.get().uri(srv.url("/test/unknown/")).finish().unwrap();
  //         let response = srv.execute(request.send()).unwrap();
  //         assert_eq!(response.status(), StatusCode::NOT_FOUND);
  //     }

  //     #[actix_rt::test]
  //     fn integration_percent_encoded() {
  //         let mut srv = test::TestServer::with_factory(|| {
  //             App::new().handler(
  //                 "test",
  //                 Files::new(".").index_file("Cargo.toml"),
  //             )
  //         });

  //         let request = srv
  //             .get()
  //             .uri(srv.url("/test/%43argo.toml"))
  //             .finish()
  //             .unwrap();
  //         let response = srv.execute(request.send()).unwrap();
  //         assert_eq!(response.status(), StatusCode::OK);
  //     }
}
