use actix_web::{
  dev::{Payload, ServiceRequest},
  http::StatusCode,
  FromRequest, HttpRequest, ResponseError,
};
use futures_util::future::{ready, Ready};
use snafu::Snafu;
use std::{
  convert::{TryFrom, TryInto},
  ops::Deref,
  path::{Path, PathBuf},
};

#[derive(Snafu, Debug, PartialEq)]
pub enum Error {
  /// The segment started with the wrapped invalid character.
  #[snafu(display("The segment started with the wrapped invalid character."))]
  BadStart { char: char },
  /// The segment contained the wrapped invalid character.
  #[snafu(display("The segment contained the wrapped invalid character."))]
  BadChar { char: char },
  /// The segment ended with the wrapped invalid character.
  #[snafu(display("The segment ended with the wrapped invalid character."))]
  BadEnd { char: char },
}

/// Return `BadRequest` for `Error`.
impl ResponseError for Error {
  fn status_code(&self) -> StatusCode {
    StatusCode::BAD_REQUEST
  }
}

type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(Debug)]
pub struct UriPathBuf(PathBuf);

impl UriPathBuf {
  pub fn new(path: &str) -> Result<Self> {
    let mut buf = PathBuf::new();

    for segment in path.split('/') {
      if segment == ".." {
        buf.pop();
      } else if segment.starts_with('.') {
        return Err(Error::BadStart { char: '.' });
      } else if segment.starts_with('*') {
        return Err(Error::BadStart { char: '*' });
      } else if segment.ends_with(':') {
        return Err(Error::BadEnd { char: ':' });
      } else if segment.ends_with('>') {
        return Err(Error::BadEnd { char: '>' });
      } else if segment.ends_with('<') {
        return Err(Error::BadEnd { char: '<' });
      } else if segment.is_empty() {
        continue;
      } else if cfg!(windows) && segment.contains('\\') {
        return Err(Error::BadChar { char: '\\' });
      } else {
        buf.push(segment)
      }
    }

    Ok(UriPathBuf(buf))
  }
}

impl FromRequest for UriPathBuf {
  type Error = Error;
  type Future = Ready<Result<Self, Self::Error>>;
  type Config = ();

  fn from_request(req: &HttpRequest, _: &mut Payload) -> Self::Future {
    ready(req.try_into())
  }
}

impl TryFrom<&HttpRequest> for UriPathBuf {
  type Error = Error;

  fn try_from(request: &HttpRequest) -> Result<Self, Error> {
    UriPathBuf::new(request.match_info().path())
  }
}

impl TryFrom<&ServiceRequest> for UriPathBuf {
  type Error = Error;

  fn try_from(request: &ServiceRequest) -> Result<Self, Error> {
    UriPathBuf::new(request.match_info().path())
  }
}

impl From<UriPathBuf> for PathBuf {
  fn from(buf: UriPathBuf) -> PathBuf {
    buf.0
  }
}

impl AsRef<Path> for UriPathBuf {
  fn as_ref(&self) -> &Path {
    self.0.deref()
  }
}

#[cfg(test)]
mod tests {
  use super::{Error, UriPathBuf};
  use std::{iter::FromIterator, path::PathBuf};

  #[actix_rt::test]
  async fn test_path_buf() {
    assert_eq!(
      UriPathBuf::new("/test/.tt").map(|t| t.0),
      Err(Error::BadStart { char: '.' })
    );
    assert_eq!(
      UriPathBuf::new("/test/*tt").map(|t| t.0),
      Err(Error::BadStart { char: '*' })
    );
    assert_eq!(
      UriPathBuf::new("/test/tt:").map(|t| t.0),
      Err(Error::BadEnd { char: ':' })
    );
    assert_eq!(
      UriPathBuf::new("/test/tt<").map(|t| t.0),
      Err(Error::BadEnd { char: '<' })
    );
    assert_eq!(
      UriPathBuf::new("/test/tt>").map(|t| t.0),
      Err(Error::BadEnd { char: '>' })
    );
    assert_eq!(
      UriPathBuf::new("/seg1/seg2/").unwrap().0,
      PathBuf::from_iter(vec!["seg1", "seg2"])
    );
    assert_eq!(
      UriPathBuf::new("/seg1/../seg2/").unwrap().0,
      PathBuf::from_iter(vec!["seg2"])
    );
  }
}
